HUDStudCount = HUDStudCount or class()

HUDStudCount.stud_icons = {
	silver = {
		texture = "units/pd2_mod_lego/particles/studs",
		texture_rect = {0, 0, 64, 64},
		animated = true,
		uv_anim_jump = {64, 0},
		uv_anim_frames = 32,
		uv_anim_framerate = 30
	},
	gold = {
		texture = "units/pd2_mod_lego/particles/studs",
		texture_rect = {0, 64, 64, 64},
		animated = true,
		uv_anim_jump = {64, 0},
		uv_anim_frames = 32,
		uv_anim_framerate = 30
	},
	blue = {
		texture = "units/pd2_mod_lego/particles/studs",
		texture_rect = {0, 128, 64, 64},
		animated = true,
		uv_anim_jump = {64, 0},
		uv_anim_frames = 32,
		uv_anim_framerate = 30
	},
	purple = {
		texture = "units/pd2_mod_lego/particles/studs",
		texture_rect = {0, 192, 64, 64},
		animated = true,
		uv_anim_jump = {64, 0},
		uv_anim_frames = 32,
		uv_anim_framerate = 30
	}
}

HUDStudCount.gui_data = {
	_general = {
		bounce_multiplier = 1.5,
		bounce_up_speed = 7,
		bounce_down_speed = 5,
		font = "fonts/font_lego_star_wars",
		text_color = Color("F4BB09")
	},
	stud_icon = {
		x = 35,
		y = 70,
		base_size = 40,
		default_icon = "silver"
	},
	stud_count = {
		x = 60,
		y = 70,
		base_size = 30
	},
	floating_stud = {	
		base_size = 110,
		target_size = 40,
		lerp_time = 0.3
	},
	true_heister_bar = {
		texture = "guis/lego_studs/true_heister",
		icon_size = {64, 64},
		frames = 10,
		stud_count = 10,
		base_size = 40,
		stud_offset = 11,
		wave_speed = 10,
		big_bounce_down_time = 0.5,
		big_bounce_down_multiplier = 2,

		drop_down_top_y = -20,
		drop_down_bottom_y = 80,

		drop_down_down_time = 0.3,
		drop_down_hold_time = 1.5,
		drop_down_up_time = 0.6
	},
	true_heister_text = {
		base_size = 150,
		shrink_size = 85,
		grow_size = 205,
		text_id = "hud_true_heister",
		lerp_time = 1
	}
}

function HUDStudCount:init(safe_hud, fullscreen_hud, sound_source)
	self._hud_panel = safe_hud.panel
	self._fullscreen_panel = fullscreen_hud.panel
	self._internal_value = 0

	self._sound_source = sound_source

	if self._hud_panel:child("stud_count_panel") then
		self._hud_panel:remove(self._hud_panel:child("stud_count_panel"))
	end

	self._stud_count_panel = self._hud_panel:panel({
		name = "stud_count_panel"
	})

	self._stud_icon = self._stud_count_panel:bitmap({
		name = "stud_icon",
		layer = 0
	})
	self:_set_stud_icon(self._stud_icon, self.gui_data.stud_icon.default_icon)

	local font = self.gui_data._general.font
	local text_color = self.gui_data._general.text_color

	self._stud_count = self._stud_count_panel:text({
		name = "stud_count",
		vertical = "center",
		text = "0",
		font = font,
		color = text_color,
		layer = 1
	})

	self._true_heister_stud_panel = self._fullscreen_panel:panel({
		name = "true_heister_stud_panel",
		height = self.gui_data.true_heister_bar.big_bounce_down_multiplier * self.gui_data.true_heister_bar.base_size,
		y = 100
	})

	self._true_heister_studs = {}
	for i = 1, self.gui_data.true_heister_bar.stud_count do
		self._true_heister_studs[i] = self._true_heister_stud_panel:bitmap({
			width = self.gui_data.true_heister_bar.base_size,
			height = self.gui_data.true_heister_bar.base_size,
			texture = self.gui_data.true_heister_bar.texture
		})
	end

	local true_heister_text = managers.localization:text(self.gui_data.true_heister_text.text_id)

	self._true_heister_text_grow = self._stud_count_panel:text({
		name = "true_heister_text_grow",
		align = "center",
		vertical = "center",
		text = true_heister_text,
		font = font,
		color = text_color,
		layer = 1
	})

	self._true_heister_text_shrink = self._stud_count_panel:text({
		name = "true_heister_text_shrink",
		align = "center",
		vertical = "center",
		text = true_heister_text,
		font = font,
		color = text_color,
		layer = 0
	})

	self:_set_bounce_scale(1)
	self:_set_true_heister_bar_percentage(0)
	self:_set_true_heister_bar_drop_lerp(0)

	self:_set_true_heister_text_lerp(1)

	self._true_heister_active = false
	self._true_heister_value = 0

	self._true_heister_current_bar = 0
	self._true_heister_bar_return_time = 0
	self._true_heister_bar_anim_thread = nil

	self._current_bounce = 0
	self._bounce_anim_thread = nil
end

function HUDStudCount:set_true_heister_value(value, already_completed)
	self._true_heister_value = value

	if already_completed then
		self:_set_true_heister_active(true)
	end
end

function HUDStudCount:_check_true_heister_status()
	if self._internal_value >= self._true_heister_value then
		self:_set_true_heister_active(false)
	else
		self:_set_true_heister_bar_percentage(self._internal_value/self._true_heister_value)
	end
end

-- Bounce
	function HUDStudCount:_set_bounce_scale(scale)
		self:_set_stud_icon_scale(scale)
		self:_set_stud_count_scale(scale)

		if not self._true_heister_active then
			self:_set_true_heister_bar_scale(scale)
		end
	end

	function HUDStudCount:_set_stud_icon_scale(scale)
		local origin_x = self.gui_data.stud_icon.x
		local origin_y = self.gui_data.stud_icon.y

		local size = self.gui_data.stud_icon.base_size * scale

		self._stud_icon:set_size(size, size)
		self._stud_icon:set_center(origin_x, origin_y)
	end

	function HUDStudCount:_set_stud_count_scale(scale)
		local origin_x = self.gui_data.stud_count.x
		local origin_y = self.gui_data.stud_count.y

		local font_size = self.gui_data.stud_count.base_size * scale

		self._stud_count:set_height(font_size)
		self._stud_count:set_font_size(font_size)
		self._stud_count:set_center_y(origin_y)
		self._stud_count:set_left(origin_x)
	end

	function HUDStudCount:_set_true_heister_bar_scale(scale)
		local origin_x = self._true_heister_stud_panel:w() / 2
		local origin_y = self._true_heister_stud_panel:h() / 2

		local size = self.gui_data.true_heister_bar.base_size * scale
		local offset = self.gui_data.true_heister_bar.stud_offset * scale

		local half_width = ((#self._true_heister_studs * offset) + size - offset) / 2

		local stud_left_x = 0
		for index, true_heister_stud in pairs(self._true_heister_studs) do
			stud_left_x = (origin_x - half_width) + ((index - 1) * offset)

			true_heister_stud:set_size(size, size)
			true_heister_stud:set_center_y(origin_y)
			true_heister_stud:set_left(stud_left_x)
		end
	end

-- True Heister Bar
	function HUDStudCount:_set_true_heister_bar_drop_lerp(lerp)
		local top = self.gui_data.true_heister_bar.drop_down_top_y
		local bottom = self.gui_data.true_heister_bar.drop_down_bottom_y

		local origin_y = math.lerp(top, bottom, lerp)

		self._true_heister_stud_panel:set_center_y(origin_y)
		self._true_heister_stud_panel:set_center_x(self._fullscreen_panel:center_x())
	end

	function HUDStudCount:_bump_true_heister_bar()
		local function bump_anim(panel)
			local down_speed = 1/self.gui_data.true_heister_bar.drop_down_down_time
			local up_speed = 1/self.gui_data.true_heister_bar.drop_down_up_time

			local going_up = false

			local lerp = 1
			local t = 0
			local dt = 0

			local going_up_time = self.gui_data.true_heister_bar.drop_down_hold_time

			while true do
				if not alive(panel) then
					return
				end

				dt = coroutine.yield()
				t = t + dt

				if self._true_heister_current_bar >= 1 and t > going_up_time then
					going_up = true
				end

				if going_up then
					self._true_heister_current_bar = self._true_heister_current_bar - (up_speed * dt)
				elseif self._true_heister_current_bar < 1 then
					self._true_heister_current_bar = self._true_heister_current_bar + (down_speed * dt)
				else
					self._true_heister_current_bar = 1
				end

				self:_set_true_heister_bar_drop_lerp(self._true_heister_current_bar)

				if self._true_heister_current_bar <= 0 and going_up == true then
					return
				end
			end
		end

		if self._true_heister_bar_anim_thread then
			self._true_heister_stud_panel:stop(self._true_heister_bar_anim_thread)
		end
		self._true_heister_bar_anim_thread = self._true_heister_stud_panel:animate(bump_anim)
	end

	function HUDStudCount:_set_true_heister_bar_wave_t(t)
		local frames = self.gui_data.true_heister_bar.frames
		local wave_speed = self.gui_data.true_heister_bar.wave_speed

		local offset = math.floor(wave_speed * t)

		local icon_width = self.gui_data.true_heister_bar.icon_size[1]
		local icon_height = self.gui_data.true_heister_bar.icon_size[2]

		local frame = 0

		for index, true_heister_stud in pairs(self._true_heister_studs) do
			frame = ((index - 1) - offset) % (self.gui_data.true_heister_bar.frames - 1)

			true_heister_stud:set_texture_rect(icon_width * frame, 0, icon_width, icon_height)
		end
	end

	function HUDStudCount:_set_true_heister_bar_percentage(percentage)
		local per_stud_percentage = 1 / #self._true_heister_studs

		local frames = self.gui_data.true_heister_bar.frames
		local icon_width = self.gui_data.true_heister_bar.icon_size[1]
		local icon_height = self.gui_data.true_heister_bar.icon_size[2]

		local stud_percentage, frame = 0

		for _, true_heister_stud in ipairs(self._true_heister_studs) do
			stud_percentage = math.clamp(percentage / per_stud_percentage, 0, 1)
			percentage = percentage - per_stud_percentage

			frame = math.floor(stud_percentage * (frames - 1))

			true_heister_stud:set_texture_rect(icon_width * frame, 0, icon_width, icon_height)
		end
	end

-- True Heister Text
	function HUDStudCount:_set_true_heister_text_lerp(lerp)
		local origin_y = self._stud_count_panel:h() / 2

		local base_size = self.gui_data.true_heister_text.base_size
		local shrink_size = self.gui_data.true_heister_text.shrink_size
		local grow_size = self.gui_data.true_heister_text.grow_size

		local grow_lerped = math.lerp(base_size, grow_size, lerp)
		local shrink_lerped = math.lerp(base_size, shrink_size, lerp)

		self._true_heister_text_grow:set_height(grow_lerped)
		self._true_heister_text_grow:set_font_size(grow_lerped)
		self._true_heister_text_grow:set_center_y(origin_y)
		self._true_heister_text_grow:set_alpha(1 - lerp)

		self._true_heister_text_shrink:set_height(shrink_lerped)
		self._true_heister_text_shrink:set_font_size(shrink_lerped)
		self._true_heister_text_shrink:set_center_y(origin_y)
		self._true_heister_text_shrink:set_alpha(1 - lerp)
	end

function HUDStudCount:_set_true_heister_active(silent)
	if self._true_heister_active then return end
	self._true_heister_active = true

	if not silent then
		self._sound_source:post_event("true_heister")
	end

	self._stud_count_panel:animate(function(panel)
		local t, dt = 0

		while true do
			if not alive(panel) then
				return
			end

			dt = coroutine.yield()
			t = t + dt

			self:_set_true_heister_bar_wave_t(t)
		end
	end)

	self._stud_count_panel:animate(function(panel)
		local lerp_speed = 1/self.gui_data.true_heister_text.lerp_time

		local t, dt = 0
		local lerp = 0

		while true do
			if not alive(panel) then
				return
			end

			dt = coroutine.yield()
			t = t + dt

			lerp = t * lerp_speed

			self:_set_true_heister_text_lerp(lerp)

			if lerp >= 1 then
				self:_set_true_heister_text_lerp(1)
				return
			end
		end
	end)

	self._stud_count_panel:animate(function(panel)
		local big_bounce_down_speed = 1/self.gui_data.true_heister_bar.big_bounce_down_time

		local t, dt = 0
		local lerp = 0
		local scale = self.gui_data.true_heister_bar.big_bounce_down_multiplier

		while true do
			if not alive(panel) then
				return
			end

			dt = coroutine.yield()
			t = t + dt

			lerp = t * big_bounce_down_speed
			scale = math.lerp(self.gui_data.true_heister_bar.big_bounce_down_multiplier, 1, lerp)

			self:_set_true_heister_bar_scale(scale)

			if lerp >= 1 then
				self:_set_true_heister_bar_scale(1)
				return
			end
		end
	end)
end

function HUDStudCount:_set_stud_icon(bitmap, id)
	local stud_icon_data = self.stud_icons[id]

	local texture = stud_icon_data.texture
	local texture_rect = stud_icon_data.texture_rect

	bitmap:set_image(texture, texture_rect[1], texture_rect[2], texture_rect[3], texture_rect[4])

	if stud_icon_data.animated then
		bitmap:animate(function(icon)
			local set_texture_rect = icon.set_texture_rect

			local x_origin = texture_rect[1]
			local y_origin = texture_rect[2]
			local width = texture_rect[3]
			local height = texture_rect[4]

			local x_jump = stud_icon_data.uv_anim_jump[1]
			local y_jump = stud_icon_data.uv_anim_jump[2]

			local frames = stud_icon_data.uv_anim_frames
			local framerate = stud_icon_data.uv_anim_framerate

			local t, dt = 0
			local frame = 0

			local x_t_jump, y_t_jump = 0

			while true do
				if not alive(icon) then
					return
				end

				dt = coroutine.yield()
				t = t + dt

				frame = math.floor((t * framerate) % frames)
				x_t_jump = x_jump * frame
				y_t_jump = y_jump * frame

				set_texture_rect(icon, x_origin + x_t_jump, y_origin + y_t_jump, width, height)
			end
		end)
	end
end

function HUDStudCount:spawn_stud(icon, value, position)
	self:_bump_true_heister_bar()

	local floating_stud = self._hud_panel:bitmap({
		width = self.gui_data.floating_stud.base_size,
		height = self.gui_data.floating_stud.base_size
	})

	self:_set_stud_icon(floating_stud, icon)
	floating_stud:set_center(position.x, position.y)

	floating_stud:animate(function(stud)
		local set_center = stud.set_center
		local set_size = stud.set_size

		local x_origin = position.x
		local y_origin = position.y

		local x_target = self._stud_count_panel:x() + self._stud_icon:center_x()
		local y_target = self._stud_count_panel:y() + self._stud_icon:center_y()

		local size_origin = self.gui_data.floating_stud.base_size
		local size_target = self.gui_data.floating_stud.target_size

		local lerp_speed = 1/self.gui_data.floating_stud.lerp_time

		local t, dt = 0
		local lerp = 0

		local x_lerped = x_origin
		local y_lerped = y_origin
		local size_lerped = size_origin

		while true do
			if not alive(stud) then
				return
			end

			dt = coroutine.yield()
			t = t + dt

			lerp = t * lerp_speed

			x_lerped = math.lerp(x_origin, x_target, lerp)
			y_lerped = math.lerp(y_origin, y_target, lerp)
			size_lerped = math.lerp(size_origin, size_target, lerp)

			set_center(stud, x_lerped, y_lerped)
			set_size(stud, size_lerped, size_lerped)

			if lerp >= 1 then
				self._hud_panel:remove(stud)
				self:add_value(icon, value)				

				return
			end
		end
	end)
end

function HUDStudCount:add_value(icon, value)
	self:_set_stud_icon(self._stud_icon, icon)

	self._sound_source:post_event("stud_counter")

	self:set_stud_count(self._internal_value + value)

	local function bounce_anim(panel)
		local bounce_up_speed = self.gui_data._general.bounce_up_speed
		local bounce_down_speed = self.gui_data._general.bounce_down_speed

		local bouncing_up = true

		local scale_lerped = 1
		local dt = 0

		while true do
			if not alive(panel) then
				return
			end

			dt = coroutine.yield()

			if self._current_bounce >= 1 then
				bouncing_up = false
			end

			if bouncing_up then
				self._current_bounce = self._current_bounce + (bounce_up_speed * dt)
			else
				self._current_bounce = self._current_bounce - (bounce_down_speed * dt)
			end

			scale_lerped = math.lerp(1, self.gui_data._general.bounce_multiplier, self._current_bounce)

			self:_set_bounce_scale(scale_lerped)

			if self._current_bounce <= 0 and bouncing_up == false then
				return
			end
		end
	end

	if self._bounce_anim_thread then
		self._stud_count_panel:stop(self._bounce_anim_thread)
	end
	self._bounce_anim_thread = self._stud_count_panel:animate(bounce_anim)
end

function HUDStudCount:set_stud_count(count)
	self._internal_value = count
	self._stud_count:set_text(managers.money:add_decimal_marks_to_string(tostring(self._internal_value)))
	self:_check_true_heister_status()
end