StudManager = StudManager or class()

StudManager.stud_info = {
	studs = {
		silver = {
			effect = Idstring("units/pd2_mod_lego/particles/silver_stud"),
			value = 10,
			bounce_sound = "stud_bounce",
			collect_sound = "stud_collect_silver",
			icon = "silver"
		},
		gold = {
			effect = Idstring("units/pd2_mod_lego/particles/gold_stud"),
			value = 100,
			bounce_sound = "stud_bounce",
			collect_sound = "stud_collect_gold",
			icon = "gold"
		},
		blue = {
			effect = Idstring("units/pd2_mod_lego/particles/blue_stud"),
			value = 1000,
			bounce_sound = "stud_bounce",
			collect_sound = "stud_collect_blurple",
			icon = "blue"
		},
		purple = {
			effect = Idstring("units/pd2_mod_lego/particles/purple_stud"),
			value = 10000,
			bounce_sound = "stud_bounce",
			collect_sound = "stud_collect_blurple",
			icon = "purple"
		}
	},

	value_order = {
		"purple",
		"blue",
		"gold",
		"silver"
	},

	stud_base_duration = 6.75,
	stud_duration_random_extra = 0.25,
	stud_duration_blink_time = 2,
	stud_range_multiplier = 1.25,
	stud_bounces = 5,
	stud_bounce_force = 550,
	stud_bounce_force_random_extra = 50,
	stud_drag = 0.8,
	stud_gravity = Vector3(0, 0, -981),

	stud_collection_offset = Vector3(0, 0, 30),
	stud_collection_distance_sq = 100 * 100,
	stud_collection_attraction_force = 300,
	stud_collection_attraction_distance_sq = 300 * 300
}

function StudManager:init()
	self._effect_manager = World:effect_manager()

	self._stud_slotmask = managers.slot:get_mask("world_geometry", "vehicles")

	self._studs_wanted = {}
	self._studs = {}

	self._source = SoundDevice:create_source("studs")
end

function StudManager:init_finalize()
	managers.hud:set_true_heister_value(managers.stud_storage:get_true_heister_value(), managers.stud_storage:get_true_heister_for_current_job())
	managers.hud:set_stud_count(managers.stud_storage:get_studs("heist"))
end

function StudManager:_calculate_drops(value)
	local drops = {}

	for _, stud_id in ipairs(self.stud_info.value_order) do
		local stud_value = self.stud_info.studs[stud_id].value
		local to_drop = 0

		while value >= stud_value do
			to_drop = to_drop + 1
			value = value - stud_value
		end

		drops[stud_id] = to_drop
	end


	return drops
end

local mvec3_add = mvector3.add
local mvec3_copy = mvector3.copy
local mvec3_direction = mvector3.direction
local mvec3_distance_sq = mvector3.distance_sq
local mvec3_multiply = mvector3.multiply
local mvec3_normalize = mvector3.normalize
local mvec3_set = mvector3.set
local mvec3_set_x = mvector3.set_x
local mvec3_set_y = mvector3.set_y
local mvec3_set_z = mvector3.set_z
local mvec3_subtract = mvector3.subtract
local mvec3_z = mvector3.z

local function randomOneMinusOne()
	return (math.random() * 2) - 1
end

function StudManager:spawn_studs(value, position, velocity_multiplier, velocity_direction)
	local velocity_multiplier = velocity_multiplier or 1
	local velocity_direction = velocity_direction or Vector3(0, 0, 1)

	local studs_wanted = self._studs_wanted
	local drops_wanted = self:_calculate_drops(value)

	-- log("Spawning Studs - " .. tostring(value) .. ":")
	-- log("\tSilver: " .. tostring(drops_wanted["silver"]))
	-- log("\tGold: " .. tostring(drops_wanted["gold"]))
	-- log("\tBlue: " .. tostring(drops_wanted["blue"]))
	-- log("\tPurple: " .. tostring(drops_wanted["purple"]))

	for stud_id, drop_count in pairs(drops_wanted) do
		for i = 1, drop_count do
			local velocity = velocity_direction * 5
			local spread = Vector3(randomOneMinusOne(), randomOneMinusOne(), randomOneMinusOne())

			mvec3_normalize(spread)
			mvec3_add(velocity, spread)

			mvec3_normalize(velocity)
			mvec3_multiply(velocity, self.stud_info.stud_bounce_force + (math.random() * self.stud_info.stud_bounce_force_random_extra))

			studs_wanted[#studs_wanted + 1] = {
				stud_id = stud_id,
				position = mvec3_copy(position),
				velocity = velocity,
				velocity_multiplier = velocity_multiplier
			}
		end
	end
end

local stud_effect_rotation = Rotation()
local temp_move_vector = Vector3()
local temp_gravity_vector = Vector3()
function StudManager:update(t, dt)
	local local_player = managers.player:local_player()
	local player_sound = local_player and local_player:sound()
	local player_position = local_player and (local_player:position() + self.stud_info.stud_collection_offset)

	local effect_manager = self._effect_manager

	local needs_cleaning = false
	local studs = self._studs
	local studs_wanted = self._studs_wanted

	local move_vec = temp_move_vector
	local gravity_vec = temp_gravity_vector

	mvec3_set(gravity_vec, self.stud_info.stud_gravity)
	mvec3_multiply(gravity_vec, dt)

	for _, wanted_data in pairs(studs_wanted) do
		local stud_data = self.stud_info.studs[wanted_data.stud_id]

		local stud_effect_id = effect_manager:spawn({
			effect = stud_data.effect,
			position = wanted_data.position,
			rotation = stud_effect_rotation
		})

		local duration = self.stud_info.stud_base_duration + (math.random() * self.stud_info.stud_duration_random_extra)
		studs[#studs + 1] = {
			lifetime = 0,
			effect_id = stud_effect_id,
			value = stud_data.value,
			position = wanted_data.position,
			velocity = wanted_data.velocity * wanted_data.velocity_multiplier,
			bounce_multiplier = wanted_data.velocity_multiplier,
			duration = duration,
			blink_time = duration - self.stud_info.stud_duration_blink_time,
			hidden = false,
			frozen = false,
			bounces = self.stud_info.stud_bounces,
			bounce_sound = stud_data.bounce_sound,
			collect_sound = stud_data.collect_sound,
			icon = stud_data.icon
		}
	end

	self._studs_wanted = {}

	local stud_entry, lifetime, duration, blink_time, effect_id, hidden = nil

	for i = 1, #studs do
		stud_entry = studs[i]

		lifetime = stud_entry.lifetime + dt
		duration = stud_entry.duration
		blink_time = stud_entry.blink_time
		effect_id = stud_entry.effect_id
		hidden = stud_entry.hidden

		if lifetime < duration then
			stud_entry.lifetime = lifetime

			local ignore_horizontal_collision = false
			local ignore_vertical_collision = false
			local collected = false

			if local_player then
				local distance_to_player_sq = mvec3_distance_sq(stud_entry.position, player_position)
				if distance_to_player_sq < self.stud_info.stud_collection_attraction_distance_sq then
					mvec3_direction(move_vec, stud_entry.position, player_position)
					mvec3_multiply(move_vec, self.stud_info.stud_collection_attraction_force * dt)
					mvec3_add(stud_entry.velocity, move_vec)

					ignore_horizontal_collision = true

					if distance_to_player_sq < self.stud_info.stud_collection_distance_sq then
						managers.stud_storage:add_studs("heist", stud_entry.value)
						managers.hud:add_studs(stud_entry.icon, stud_entry.value, stud_entry.position)

						if i <= 32 then -- Limit the amount of sounds to avoid OOMing.
							player_sound:play(stud_entry.collect_sound, nil, false)
						end

						effect_manager:kill(effect_id)

						studs[i] = false
						needs_cleaning = true

						collected = true
					end
				end
			end

			if not collected then
				mvec3_set(move_vec, stud_entry.velocity)
				mvec3_multiply(move_vec, (1 - self.stud_info.stud_drag) * dt)
				mvec3_subtract(stud_entry.velocity, move_vec)

				mvec3_set(move_vec, stud_entry.velocity)
				mvec3_multiply(move_vec, dt)
				mvec3_add(move_vec, stud_entry.position)
				mvec3_set(stud_entry.position, move_vec)

				mvec3_add(stud_entry.velocity, gravity_vec)

				local col_ray = World:raycast("ray", stud_entry.position, stud_entry.position + (math.DOWN * 30), "slot_mask", self._stud_slotmask)
				if col_ray then
					if stud_entry.bounces > 0 then
						stud_entry.bounces = stud_entry.bounces - 1
						local bounce_percent = stud_entry.bounces / self.stud_info.stud_bounces
						local bounce_multiplier = stud_entry.bounce_multiplier

						mvec3_set_z(stud_entry.velocity, (self.stud_info.stud_bounce_force + (math.random() * self.stud_info.stud_bounce_force_random_extra)) * bounce_multiplier * bounce_percent)

						ignore_vertical_collision = true

						if i <= 32 then -- Limit the amount of sounds to avoid OOMing.
							self._source:set_position(stud_entry.position)
							self._source:post_event("stud_bounce")
						end
					else
						mvec3_set_z(stud_entry.velocity, 0)
						mvec3_set_z(stud_entry.position, mvec3_z(col_ray.position) + 30)
					end
				end

				effect_manager:move(effect_id, stud_entry.position)

				if not (ignore_vertical_collision and ignore_horizontal_collision) then
					local hit_bodies = World:find_bodies("intersect", "sphere", stud_entry.position, 25, self._stud_slotmask)

					if #hit_bodies > 0 then
						if not ignore_horizontal_collision then
							mvec3_set_x(stud_entry.velocity, 0)
							mvec3_set_y(stud_entry.velocity, 0)
						end

						if not ignore_vertical_collision and mvec3_z(stud_entry.velocity) > 0 then
							mvec3_set_z(stud_entry.velocity, 0)
						end
					end
				end

				if lifetime > blink_time then
					local blink_value = math.sin(lifetime * 3600)

					if (blink_value >= 0) and hidden then
						effect_manager:set_hidden(effect_id, false)
						stud_entry.hidden = false
					elseif (blink_value < 0) and not hidden then
						effect_manager:set_hidden(effect_id, true)
						stud_entry.hidden = true
					end
				end
			end
		else
			effect_manager:kill(effect_id)

			studs[i] = false
			needs_cleaning = true
		end
	end

	if needs_cleaning then
		local new_studs = {}

		for i = 1, #studs do
			stud_entry = studs[i]

			if stud_entry then
				new_studs[#new_studs + 1] = stud_entry
			end
		end

		self._studs = new_studs
	end
end

Hooks:PostHook(GameSetup, "init_managers", "InitStudManager", function(self)
	managers.stud = StudManager:new()
end)

Hooks:PostHook(GameSetup, "init_finalize", "InitFinalizeStudManager", function(self)
	managers.stud:init_finalize()
end)

Hooks:PostHook(GameSetup, "update", "UpdateStudManager", function(self, t, dt)
	managers.stud:update(t, dt)
end)