StudStorageManager = StudStorageManager or class()

function StudStorageManager:init()
	Global.studs_collected = Global.studs_collected or {
		heist = 0,
		total = 0
	}
	self._studs_collected = Global.studs_collected

	Global.true_heister_jobs = Global.true_heister_jobs or {}
	self._true_heister_jobs = Global.true_heister_jobs

	self._true_heister_value = 0
	if managers.job:current_job_chain_data() then
		self._true_heister_value = #managers.job:current_job_chain_data() * 30000
	end

	if managers.job:on_first_stage() and not managers.job:interupt_stage() then
		self:reset_heist_studs()
	end

	self._initial_heist_studs = self._studs_collected.heist
end

function StudStorageManager:get_studs(type)
	return self._studs_collected[type]
end

function StudStorageManager:add_studs(type, count)
	self._studs_collected[type] = self._studs_collected[type] + count
end

function StudStorageManager:store_heist_studs()
	self._studs_collected.total = self._studs_collected.total + self._studs_collected.heist

	self._studs_collected.heist = 0
end

function StudStorageManager:reset_heist_studs()
	self._studs_collected.heist = 0
end

function StudStorageManager:get_true_heister_value()
	return self._true_heister_value
end

function StudStorageManager:get_true_heister_for_current_job()
	local job_id = managers.job:current_job_id()
	return self:get_true_heister_for_job_id(job_id)
end

function StudStorageManager:get_true_heister_for_job_id(job_id)
	return self._true_heister_jobs[job_id] or false
end

function StudStorageManager:store_true_heister()
	local job_id = managers.job:current_job_id()

	if self._studs_collected.heist > self._true_heister_value then
		self._true_heister_jobs[job_id] = true
	end
end

function StudStorageManager:on_retry_job_stage()
	self._studs_collected.heist = self._initial_heist_studs
end

function StudStorageManager:on_heist_completed()
	self:store_true_heister()
	self:store_heist_studs()
end

function StudStorageManager:save(data)
	local state  = {
		studs_collected = clone(self._studs_collected),
		true_heister_jobs = clone(self._true_heister_jobs)
	}

	data.stud_storage = state
end

function StudStorageManager:load(data)
	local state = data.stud_storage

	if state then
		self._studs_collected = state.studs_collected
		Global.studs_collected = self._studs_collected

		self._true_heister_jobs = state.true_heister_jobs
		Global.true_heister_jobs = self._true_heister_jobs
	end
end

Hooks:PostHook(Setup, "init_managers", "InitStudStorageManager", function(self, managers)
	managers.stud_storage = StudStorageManager:new()
end)

Hooks:PostHook(Setup, "save", "SaveStudStorageManager", function(self, data)
	managers.stud_storage:save(data)
end)

Hooks:PostHook(Setup, "load", "LoadStudStorageManager", function(self, data)
	managers.stud_storage:load(data)
end)

Hooks:PostHook(ExperienceManager, "save", "SaveFileStudStorageManager", function(self, cache)
	managers.stud_storage:save(cache)
end)

Hooks:PostHook(ExperienceManager, "load", "LoadFileStudStorageManager", function(self, cache, version)
	managers.stud_storage:load(cache)
end)